package epe.agents;

	import java.awt.event.ActionListener;
	
	import javax.swing.Timer;
	
	import epe.gui.Background;
	import epe.gui.GUI;
	import epe.gui.ProgrammingEnvironment;
	import epe.gui.SystemPrinter;
	import epe.agents.acquaintances.Acquaintance;
	import epe.agents.acquaintances.AcquaintancesManager;
	import epe.agents.agentInterface.AgentInterface;
	import epe.agents.agentListeners.messageSend.MessageSendListener;
	import epe.agents.agentListeners.messageSend.MessageSendListeners;
	import epe.agents.agentListeners.messageSend.MessageSendNotifier;
	import epe.agents.agentListeners.ncs.NcsListener;
	import epe.agents.agentListeners.ncs.NcsListeners;
	import epe.agents.agentListeners.ncs.NcsNotifier;
	import epe.agents.agentListeners.reorganization.ReorganizationListener;
	import epe.agents.agentListeners.reorganization.ReorganizationListeners;
	import epe.agents.agentListeners.reorganization.ReorganizationNotifier;
	import epe.communication.Message;
	import epe.communication.MessageListener;
	import epe.communication.MessageManager;
	import epe.communication.MessageSender;
	import epe.communication.messageBox.MessageBox;
	import epe.communication.messages.contract
						.ProposingCooperationInWithPotentialValueMessage;
	import epe.communication.messages.contract
						.ProposingCooperationMessage;
	import epe.communication.messages.contract
						.ProposingCooperationOutWithPrejudiceMessage;
	import epe.communication.messages.ncs.NCSInformedNeedInMessage;
	import epe.communication.messages.ncs.NCSNeedInMessage;
	import epe.communication.messages.ncs.NCSNeedOutMessage;
	import epe.reorganization.ReorganizationManager;
	import epe.synchronization.SynchronizationManager;


// This is the main class for the generic activity of an agent.
// It is also the centraliser for all the different managers, functionalities and tools.
// IMPORTANT info concerning IN/OUT and the messages : the notion of IN and OUT is from 
// the requesting agent point of view. A NeedInMessage means the agent is missing some input.
// And the associated proposition message is ProposingInMessage : the other agent is
// proposing what the requesting agent needs.

public abstract class Agent extends Thread 
                            implements MessageListener,
                                       MessageSender,
                                       ActionListener,
                                       NcsNotifier,
                                       MessageSendNotifier,
                                       ReorganizationNotifier {
  
    // Agent attributes
    protected AcquaintancesManager acquaintancesManager;
    protected MessageManager messageManager;
    protected MessageBox messageBox;
    protected SynchronizationManager synchronizationManager;
    protected ReorganizationManager reorganizationManager;
    protected String name;
    protected GUI gui;
    
    // Listeners attributes
    protected NcsListeners ncsListeners;
    protected MessageSendListeners messageSendListeners;
    protected ReorganizationListeners reorganizationListeners;
    
    // The environment where the agent executes itself
    protected ProgrammingEnvironment programmingEnvironment;
    
    // Thread attributes
    protected boolean runningForTheFirstTime = true;
    protected boolean enabled = false;
    protected boolean running = false;
    protected boolean pauseSet = true;
    protected final static int ONE_SECOND = 1000;
    protected int mainDelay = 0*ONE_SECOND; //3
    protected int messageDelay = 0; //100
    protected boolean acted = false;
    protected boolean inCooperativeState = false;
    protected boolean reInitializeOrderSend = false;
    protected boolean receivedEvent = false;
  
    // Timer to tell when the agent is ready to act again
    // (usefull not to flood the environment or act faster than the reorganization process)
    protected Timer actedTimer;
    protected int actedTimerDelay = 100;//1*ONE_SECOND;//1
    protected boolean actedTimerRanOut = true;
    
    // Timer to tell when the agent is ready to receive data
    protected boolean repeatReadyToReceive = false;
    protected Timer sendReadyToReceiveTimer;
    protected int sendReadyToReceiveTimerDelay = 1*ONE_SECOND;//5
    protected boolean sendReadyToReceiveTimerRanOut = true;
    
    // Timer to detect blocking situations
    protected Timer blockingSituationTimer;
    protected int blockingSituationTimerDelay = 2*ONE_SECOND;//5
    protected boolean blockingSituationTimerRanOut = true;
    protected boolean longTimeNotActed = true;
    protected boolean blockingSituationNeedsSolving = false;
    // needed for solving : commutating between two possibilities
    protected boolean blockingSituationCommutator = true; 
    
    // Output options
    protected boolean debug = true;
    protected boolean verbose = true;
    protected boolean extendedVerbose = false;
    protected boolean agentTextOutputEnabled = true;
     
    // Counter for the agent to judge if he is still in a reorganization phase 
    // (higher/lower reorganization)
    protected int reorganizationPhaseCounter = 0;
    protected int reorganizationPhaseCounterThreshold = 10;
    

  public Agent(String name) {
    this.name = name;
    gui = new GUI(this);
    messageManager = new MessageManager(this);
    messageBox = new MessageBox();
    ncsListeners = new NcsListeners();
    messageSendListeners = new MessageSendListeners();
    reorganizationListeners = new ReorganizationListeners();
    sendReadyToReceiveTimer = new Timer(sendReadyToReceiveTimerDelay, this);
    sendReadyToReceiveTimer.setRepeats(false);
    blockingSituationTimer = new Timer(blockingSituationTimerDelay, this);
    blockingSituationTimer.setRepeats(false);
    actedTimer = new Timer(actedTimerDelay, this);
    actedTimer.setRepeats(false);
  }
  
  public void setProgrammingEnvironment(ProgrammingEnvironment programmingEnvironment) {
    this.programmingEnvironment = programmingEnvironment;
    gui.showAgentGUIinEnvironment();
    // GraphLayoutTool notification and agent initialisation
    addReorganizationListener(getProgrammingEnvironment().getGraphLayoutTool());
    addReorganizationListener(getProgrammingEnvironment().getLinkObserver());
    graphLayoutToolAgentInitialisation();
  }
  // Specific to each agent because of its type (different representation in the GraphLayoutTool)
  protected abstract void graphLayoutToolAgentInitialisation();
    
  public void setLocation(int xLocation, int yLocation) {
    gui.setLocation(xLocation, yLocation, 0, 0);
    programmingEnvironment.locationChanged();
  }
  
  public void setLocation(int xLocation, int yLocation, int shiftX, int shiftY) {
    gui.setLocation(xLocation, yLocation, shiftX, shiftY);
    programmingEnvironment.locationChanged();
  }
  
  public synchronized void reInitializeOrder() {
    if (extendedVerbose) gui.println(name + " : reInitializeOrder send.", GUI.green);
    reInitializeOrderSend = true;
    notifyAll();
  }
  
  public void reInitializeAgent() {
    gui.println("REINITIALIZATION", GUI.pink);
    synchronizationManager.reInitialize();
    reorganizationManager.reInitialize();
    reInitializeBlockingSituationManager();
    gui.println("");
    gui.println("");
    gui.println("");
    
  }
  
// Run loop for the Thread
  public void run() { 
    if (extendedVerbose) gui.println(name + " : run method called", GUI.green);
    while (true) {
      
      if (reInitializeOrderSend) {
        reInitializeOrderSend = false;
        reInitializeAgent();
      }
      
      while (running && enabled) {
      	gui.println("-------------------------------------------------------------------------");
        if (extendedVerbose) gui.println("RUN loop phase start : " + 
        																 name + " running", GUI.green);
        initialDelay();
        gui.start();
        setActed(false);
        if (!isMessageBoxEmpty()){
          if (extendedVerbose) gui.println("Processing a message. ", GUI.green);
          processOneMessage();
          if (isMessageBoxEmpty())
            // When the last message is processed, the agent has to wait before firing a NCS
            // Indeed, it should wait for a possible answer
            // This is handled as if it fired a NCS
            reorganizationManager.ncsFired();
          setActed(true);
          cancelLongTimeNotActed();
        }
        else {
          if (extendedVerbose) gui.println("MessageBox is empty. ", GUI.green);
          managingNCSs();
          if (inCooperativeState && actedTimerRanOut) {
            gui.println(name + " in Cooperative State and ready to act", GUI.yellow);
            reorganizationPhaseCounter++;
            
            acting();
            
            if (acted) {
              if (reorganizationPhaseCounter >= reorganizationPhaseCounterThreshold)
              reorganizationManager.setReorganizationPhase(false);
              fireActedTimer();
            }
          }
          else {
            reorganizationPhaseCounter = 0;
            if (!actedTimerRanOut)
            	gui.println("Not ready to act.");
          }
        }
        pause(mainDelay);
        synchronized(this) {
          if (!hasActed() && !receivedEvent) {
            stopAgent();
          }
          setReceivedEvent(false);
        }
      }

      synchronized(this) {
        if (!running | !enabled)
          try {
              if (extendedVerbose) gui.println(name + " waiting", GUI.green);
              gui.stop();
              this.wait();
          } catch(InterruptedException e) {}
      }
      //yield();
    }
  }

  // Make the agent pause
  protected void pause(int delay) {
    if (extendedVerbose) gui.println(name + " going to pause for : " 
                                          + delay + "ms.", GUI.green);
     if (pauseSet)
        try {
          gui.stop();
          sleep(delay);
          gui.start();
        } catch (InterruptedException e) {
            if (extendedVerbose) gui.println(name + " sleep interrupted", GUI.green);
          }
  }
  
  protected void initialDelay() {
    if (runningForTheFirstTime) {
      int randomDelay = (int)(java.lang.Math.random()*2*ONE_SECOND);
      if (extendedVerbose) gui.println(name + " going to wait for : " + randomDelay, 
                               GUI.green);
      //pause(randomDelay);
      runningForTheFirstTime = false;
    }
  }
  
  // Here the agent checks if there are some NCSs
  // and tries to reorganize with the other agents if needed
  protected void managingNCSs() {
    reorganizationManager.manageNCSs();
  }
  
  public void detectInOutNCSs() {
    reorganizationManager.detectInOutNCSs();
  }

  public void checkCooperativeState() {
   reorganizationManager.checkCooperativeState(); 
  }
  
  protected void agentHasComputedItsFunction() {
    acquaintancesManager.increaseAllTrustValues("computedFunction");
  }
  
  // Begin of abstract methods
  
  // It is here that the agent really acts (compute its function)
  // Override this method for each agent
  protected abstract void acting();
  
  public abstract void cooperationWithInAcquaintanceMayHaveBeenCancelled(Acquaintance acq);
  
  // Specific to each agent. Reinitialize IN and readyToReceive so as to retry the calculus
  public abstract void retryCalculus();
  
  // End of abstract methods
  
  public void setActed(boolean truthValue) {
    acted = truthValue;
  }
  
  public void setReceivedEvent(boolean truthValue) {
    receivedEvent = truthValue;
  }
  
  public void setLongTimeNotActed(boolean truthValue) {
    longTimeNotActed = truthValue;
  }
  
  public void setBlockingSituationNeedsSolving(boolean truthValue) {
    blockingSituationNeedsSolving = truthValue;
  }
  
  public boolean hasActed() {
    return acted;
  }
  
  public boolean longTimeNotActed() {
    return longTimeNotActed;
  }
  
  public boolean blockingSituationNeedsSolving() {
    return blockingSituationNeedsSolving;
  }
  
  public void setMainDelay(int delay) {
    mainDelay = delay;
  }
  
  public void setMessageDelay(int delay) {
    messageDelay = delay;
  }
  
  public synchronized void startAgent() {
    running = true;
    if (extendedVerbose) gui.println(name + " : Starting.", GUI.green);
    notifyAll();
  }
  
  public synchronized void stopAgent() {
    running = false;
    if (extendedVerbose) gui.println(name + " : Stopping.", GUI.green);
  }
  
  public synchronized void enableAgent() {
    enabled = true;
    if (extendedVerbose) gui.println(name + " : Enabled.", GUI.green);
    notifyAll();
  }
  
  public synchronized void disableAgent() {
    enabled = false;
    if (extendedVerbose) gui.println(name + " : Disabled.", GUI.green);
  }
  
// Concernant l'Agent
  public String getAgentName() {
    return name;
  }
  
  public GUI getGUI() {
    return gui;
  }
  
  public ProgrammingEnvironment getProgrammingEnvironment() {
    return programmingEnvironment;
  }
  
  public Background getBackground() {
    return getProgrammingEnvironment().getEnvironmentBackground();
  }
  
  public AcquaintancesManager getAcquaintancesManager() {
    return acquaintancesManager;
  }
  
  public ReorganizationManager getReorganizationManager() {
    return reorganizationManager;
  }
  
  public int getNumberOfINs() {
    return reorganizationManager.numberOfINs();
  }  
    
  public boolean isAgentTextOutputEnabled() {
    return agentTextOutputEnabled;
  }
  
  public void setAgentTextOutputEnabled(boolean truthValue) {
    agentTextOutputEnabled = truthValue;
  }
  
  public boolean equals(Agent agent) {
    return this == agent;
  }
  
  public void setCooperativeState(boolean truthValue) {
    inCooperativeState = truthValue;
  }
  
  public boolean inCooperativeState() {
    return inCooperativeState;
  }

  public boolean isVerbose() {
    return verbose;
  }
  
  // Flag telling if the agent is (indirectly) connected to the Output
  public boolean isUsefull() {
    return reorganizationManager.isUsefull();
  }

  protected void setRepeatReadyToReceive(boolean truthValue) {
    repeatReadyToReceive = truthValue;
  }
  
  protected void readyToReceiveFired() {
    sendReadyToReceiveTimerRanOut = false;
    sendReadyToReceiveTimer.restart();
  }
  
  protected void switchBockingSituationCommutator() {
    blockingSituationCommutator = !blockingSituationCommutator;
  }
  
  
  protected void fireBlockingSituation() {
    blockingSituationTimerRanOut = false;
    blockingSituationTimer.restart();
  }
  
  protected void fireActedTimer() {
    actedTimerRanOut = false;
    actedTimer.restart();
  }
  
  protected void reInitializeBlockingSituationManager() {
    gui.println("ReInitializing BlockingSituationManager", GUI.pink);
    blockingSituationTimer.stop();
    blockingSituationTimerRanOut = false;
    blockingSituationCommutator= true;
    setBlockingSituationNeedsSolving(false);
    setLongTimeNotActed(false);
  }
  
  // Timers call this method when they get to zero
  public void actionPerformed(java.awt.event.ActionEvent evt) {
    setReceivedEvent(true);
    if ((evt.getSource()).equals(sendReadyToReceiveTimer)) {
      sendReadyToReceiveTimerRanOut = true;
      startAgent();
    } 
    else if ((evt.getSource()).equals(blockingSituationTimer)) {
      blockingSituationTimerRanOut = true;
      if (longTimeNotActed()) {
        setBlockingSituationNeedsSolving(true);
        startAgent();
      }
    }
    else if ((evt.getSource()).equals(actedTimer)) {
      actedTimerRanOut = true;
      startAgent();
    }
    else
      gui.println("DEBUG : ERROR : unknown Timer ran out "
                + evt.getSource(), GUI.red);
  }
  
  public void cancelLongTimeNotActed() {
    setLongTimeNotActed(false);
  }
  
  // To add an acquaintance to an agent
  public void addAcquaintance(Acquaintance acquaintance) {
    acquaintancesManager.addAcquaintance(acquaintance);
  }
  
  public void addAcquaintance(Agent agent, String type, int position) {
    acquaintancesManager.addAcquaintance(agent, type, position);
  }
  
  public String addAcquaintanceWithReturn(Agent agent, String type, 
                                          int myPosition, int positionOnAcq) {
    return acquaintancesManager.addAcquaintanceWithReturn(agent, type, myPosition, 
                                                          positionOnAcq);
  }
  
  public void addAcquaintance(Agent agent, String type, int position, String dataType) {
    acquaintancesManager.addAcquaintance(agent, type, position, dataType);
  }
  
  public void addAcquaintance(Agent agent, String type, int position, boolean fixedTrustValue) {
    acquaintancesManager.addAcquaintance(agent, type, position, fixedTrustValue);
  }
 
  public void addAcquaintance(Agent agent, 
                              String type, 
                              int position, 
                              String dataType, 
                              boolean fixedTrustValue) {
    acquaintancesManager.addAcquaintance(agent, type, position, dataType, fixedTrustValue);
  }

  // Concerning the MessageListener
  // React to a message coming in (note : specific messages are captured by receiveMessage
  // from the subclass) : Data messages
  // note : the method simply puts the message in the MessageBox to be processed later.
  // Exception : the CyclePingMessages need to be processed imediately
  public synchronized void receiveMessage(Message message) {
    String typeId = message.getId();
    
    if (typeId.equals(Message.READY_TO_RECEIVE_MESSAGE)) {
      synchronizationManager.processReadyMessage(message);
      if ((blockingSituationTimerRanOut || !longTimeNotActed())
          & (!blockingSituationNeedsSolving)) {
        setLongTimeNotActed(true);
        fireBlockingSituation();
      }
    }
    else if (typeId.equals(Message.CYCLE_PING_MESSAGE)) {
    	reorganizationManager.processMessage(message);
    }
    else {
      messageBox.addMessage(message);
    }
    startAgent();
  }
 
  // Called by the specific agents when they get a DataMessage
  // Usefull to make some output
  public synchronized void receivedDataMessage(Message message)  {
    String typeId = message.getId();
    Agent sender = message.getSender();
    gui.println(getAgentName() + " : received DataMessage [ " 
                + typeId + " ] from ["
                + sender.getAgentName() + " ].", GUI.lightGray);
    cancelLongTimeNotActed();
  }
  // Called by the specific agents when they get a DataMessage 
  // but it is not in a cooperative state
  // Usefull to make some output
  public synchronized void receivedDataMessageWhileNotInCooperativeState(Message message)  {
    String typeId = message.getId();
    Agent sender = message.getSender();
    gui.println("WARNING : " + getAgentName() + " : received DataMessage [ " 
                + typeId + " ] from ["
                + sender.getAgentName() 
                + " ] while not in cooperative State", GUI.red);
  }
  public boolean isMessageBoxEmpty() {
    return messageBox.isEmpty();
  }
  
  // Traitement effectif d'un message
  protected void processOneMessage() {
    Message message = messageBox.getMessage();
    String typeId = message.getId();
    Agent sender = message.getSender();
    if (verbose) gui.println(getAgentName() + " : processing message : " + typeId 
                             + " from [" + sender.getAgentName() + "]", GUI.green);
    // Add to NEUTRAL acqs if not already known
		if (!(acquaintancesManager.isANeutralAcquaintance(sender))) {
			acquaintancesManager.addAcquaintance(sender, "NEUTRAL", 0, "UNKNOWN");
		}
		
		// Update datatypes of known NEUTRAL "UNKNOWN" acqs with datatypes from messages.
		String dataType = acquaintancesManager.getAcquaintances()
											  .getAcquaintance(sender, "NEUTRAL")
											  .getDataType();
		if (dataType.equals("UNKNOWN")) {
			if (typeId.equals(Message.NCS_NEED_IN_MESSAGE)) {
				dataType = ((NCSNeedInMessage)message).getDataType();
			}
			else if (typeId.equals(Message.NCS_NEED_OUT_MESSAGE)) {
				dataType = ((NCSNeedOutMessage)message).getDataType();
			}
			else if (typeId.equals(Message.PROPOSING_COOPERATION_MESSAGE)) {
				dataType = ((ProposingCooperationMessage)message).getDataType();
			}
			else if (typeId.equals(Message.NCS_INFORMED_NEED_IN_MESSAGE)) {
				dataType = ((NCSInformedNeedInMessage)message).getDataType();
			}
			else if (typeId.equals(Message.PROPOSING_COOPERATION_OUT_WITH_PREJUDICE_MESSAGE)) {
				dataType = ((ProposingCooperationOutWithPrejudiceMessage)message).getDataType();
			}
			else if (typeId.equals(Message.PROPOSING_COOPERATION_IN_WITH_POTENTIAL_VALUE_MESSAGE)) {
				dataType = ((ProposingCooperationInWithPotentialValueMessage)message).getDataType();
			}
			acquaintancesManager.getAcquaintances()
													.getAcquaintance(sender, "NEUTRAL")
													.setDataType(dataType);
		}
		
    if ((typeId.equals(Message.NCS_NEED_IN_MESSAGE))
			||(typeId.equals(Message.NCS_NEED_OUT_MESSAGE))
			||(typeId.equals(Message.NCS_INADEQUATE_DATA_MESSAGE))
			||(typeId.equals(Message.NCS_INFORMED_INADEQUATE_DATA_MESSAGE))
			||(typeId.equals(Message.NCS_INFORMED_NEED_IN_MESSAGE))
			||(typeId.equals(Message.NCS_MISSING_DATA_MESSAGE))
    	    ||(typeId.equals(Message.NCS_NO_POTENTIAL_OUTPUT_MESSAGE))
			||(typeId.equals(Message.PROPOSING_COOPERATION_MESSAGE))
			||(typeId.equals(Message.PROPOSING_COOPERATION_OUT_WITH_PREJUDICE_MESSAGE))
			||(typeId.equals(Message.PROPOSING_COOPERATION_IN_WITH_POTENTIAL_VALUE_MESSAGE))
			||(typeId.equals(Message.REQUESTING_COOPERATION_MESSAGE))
			||(typeId.equals(Message.ACKNOWLEDGING_COOPERATION_MESSAGE))
			||(typeId.equals(Message.REFUSING_COOPERATION_MESSAGE))
			||(typeId.equals(Message.CANCELLING_COOPERATION_MESSAGE))
			||(typeId.equals(Message.DATATYPE_CHANGED_MESSAGE))
      ||(typeId.equals(Message.POTENTIAL_OUTPUT_MESSAGE))
    	||(typeId.equals(Message.GOAL_REACHED_MESSAGE))
    	||(typeId.equals(Message.LINK_MULTIPLICITY_CHANGED_MESSAGE)))
    	reorganizationManager.processMessage(message);
    else if ((typeId.equals(Message.NUMBER_MESSAGE) || typeId.equals(Message.BOOLEAN_MESSAGE)) 
              && !inCooperativeState) {
       SystemPrinter.printSystem("WARNING : " + getAgentName());
       gui.println(name + " : WARNING : Received Data while not in cooperative state from [ " 
                       + sender.getAgentName() + " ] : "
                       + typeId + ".", GUI.red);
    }
    else {
      SystemPrinter.printSystem("ERROR : " + getAgentName());
      errorHappened();
      gui.println(name + " : ERROR : received message of unknown type from [ " 
                       + sender.getAgentName() + " ] : "
                       + typeId + ".", GUI.red);
    }
  }

// Concernant le MessageSender  
  // Envoie un message sur IN,OUT ou NEUTRAL, en sp?cifiant le n? de l'entr?e/sortie
  public void fireMessage(Message message, AgentInterface agentInterface, Agent sender) {
    
    if (enabled) {
      messageManager.forwardMessage(message, 
                                    agentInterface, 
                                    sender);
      notifyMessageSendListeners();
      pause(messageDelay);
    }
  }
  
  // Envoie le message ? un agent particulier
  public void fireMessage(Message message,
                          Agent adressee,
                          String type,
                          Agent sender) {
    
    if (enabled) {
      messageManager.forwardPersonalMessage(message, 
                                            adressee,
                                            type,
                                            sender);
      notifyMessageSendListeners();
      pause(messageDelay);
    }
  }
  
  public void fireMessageToAllAcquaintances(Message message) {
    if (enabled) {
      messageManager.sendMessageToAllAcquaintances(message);
      notifyMessageSendListeners();
      pause(messageDelay);
    }
  }
  
  public void fireMessageToAllAcquaintancesButSender(Message message) {
    if (enabled) {
      messageManager.sendMessageToAllAcquaintancesButSender(message);
      notifyMessageSendListeners();
      pause(messageDelay);
    }
  }
 
// Section "NcsNotifier":
// Each time a Ncs arises, that agent notifies its listeners
  public void addNcsListener(NcsListener listener) {
    ncsListeners.addListener(listener);
  }
  
  public void removeNcsListener(NcsListener listener) {
    ncsListeners.removeListener(listener);
  }
  
  public void notifyNcsListeners(Message message) {
     //ncsListeners.notifyListeners(message);
  }
  
  public void notifyNcsListeners() {
     ncsListeners.notifyListeners();
  }
  
  public void notifyNcsListeners(String type) {
    ncsListeners.notifyListeners(type);
  }
// end of the  "NcsNotifier" section

// Section "MessageNotifier":
// Each time a Message is send, that agent notifies its listeners
  public void addMessageSendListener(MessageSendListener listener) {
    messageSendListeners.addListener(listener);
  }
  
  public void removeMessageSendListener(MessageSendListener listener) {
    messageSendListeners.removeListener(listener);
  }
    
  public void notifyMessageSendListeners(Message message) {
  }
  
  public void notifyMessageSendListeners() {
    messageSendListeners.notifyListeners();
  }
// end of the  "MessageNotifier" section

// Section "ReorganizationNotifier":
// Each time some reorganization is done, that agent notifies its listeners
  public void addReorganizationListener(ReorganizationListener listener) {
    reorganizationListeners.addListener(listener);
  }
  
  public void removeReorganizationListener(ReorganizationListener listener) {
    reorganizationListeners.removeListener(listener);
  }
  
  public void notifyReorganizationListenersOfAddLink(String type, String source, String target) {
    reorganizationListeners.notifyReorganizationListenersOfAddLink(type, source, target);
  }
  
  public void notifyReorganizationListenersOfRemoveLink(String type, String source, String target) {
    reorganizationListeners.notifyReorganizationListenersOfRemoveLink(type, source, target);
  }
// end of the  "ReorganizationNotifier" section
  
  // When there's a major Agent error, this method is called.
  public void errorHappened() {
    gui.errorNotification();
  }
      
}