package epe.observationTools.graphLayoutTool.agents;

	import java.awt.Color;

/**
 * Internal representation of an Agent
 * @author  Bastien Hofmann
 */
public class Agent {
    
    /* color of the edge of the box */
    private final Color BOX_EDGE_COLOR = Color.black;
    
    /* label to be written on the box (and indentifier of the agent) */
    public final String label;
    
    /* type of agent */
    public final int type;
    
    public boolean wellPlaced = false;
    
    /* statics of available types */
    /* normal agent*/
    public static final int NORMAL = 0;
    /* the agent is the input agent */
    public static final int INPUT = 1;
    /* the agent is the output agent */
    public static final int OUTPUT = 2;
    /* the agent is the environment agent */
    public static final int ENVIRONMENT = 3;
    
    
    /* AgentManager */
    private final AgentManager manager;
    
    /** left of the box representing the agent
     */
    private int left;
    
    /** top of the box representing the agent
     */
    private int top;
    
    /** color of the background of the box */
    public final Color color;
    
    /* set highlight to true to show a relation to another agent */
    public boolean highlight = false;
    
    /* set to true if you want the arrow and the circle to be drawn */
    /*public static final boolean DRAW_IO = false;*/
    
    /** Creates a new instance of Agent
     * @param label identifying the agent, and appearing on the box
     */
    public Agent(java.lang.String lbl, int type, java.awt.Color color, AgentManager am) {
        // initializing the values
        this.label = lbl;
        this.type = type;
        this.color = color;
        this.manager = am;
        int w = this.manager.dimension.width;
        int h = this.manager.dimension.height;
        this.left = (int)(w * Math.random());
        this.top = (int)(h * Math.random());
    }
    
    /** Returns the left in pixels of the box
     * @return Left in pixels of the agent
     */
    public int getLeft() {
        return(this.left);
    }
    
    /** Sets the new left in pixels of the box
     * @param Left in pixels of the box
     */
    public void setLeft(int l) {
        if (l < 0) {
            l = 0;
        }
        if (l + this.getWidth() > this.manager.dimension.width) {
            l = this.manager.dimension.width - this.getWidth();
        }
        this.left = l;
    }
    
    /** Returns the top of the box in pixels
     * @return Top in pixels of the agent
     */
    public int getTop() {
        return(this.top);
    }

    /** Sets the new top in pixels of the box
     * @param Top in pixels of the box
     */
    public void setTop(int t) {
        if (t < 0) {
            t = 0;
        }
        if (t + this.getHeight() > this.manager.dimension.height) {
            t = this.manager.dimension.height - this.getHeight();
        }
        this.top = t;
    }

    /** horizontal offset wanted by the organizers
     */
    public int offsetX = 0;
    
    /** vertical offset wanted by the organizers
     */
    public int offsetY = 0;
    
    /** Tells the agent the direction he will have to move to
     * @param x coordinate of the point to move the agent in direction to
     * @param y coordinate of the point to move the agent in direction to
     */
    public boolean moveTo(int x, int y) {
        int o;

        if (this.left != x) {
            o = (int)((x - this.left) * .05);
            if (o == 0) {
                if (x > this.left) {
                    this.offsetX++;
                } else {
                    this.offsetX--;
                }
            } else {
                this.offsetX += o;
            }
        }
        if (this.top != y) {
            o = (int)((y - this.top) * .05);
            if (o == 0) {
                if (y > this.top) {
                    this.offsetY++;
                } else {
                    this.offsetY--;
                }
            } else {
                this.offsetY += o;
            }
        }
        return((this.left == x) && (this.top == y));
    }
    
    /** Tells the agent to move
     */
    public void move() {
        // the agent doesn't move if he is being dragged by the user
        if (this != AgentManager.picked) {
            this.setLeft(this.left + this.offsetX);
            this.setTop(this.top + this.offsetY);
        }
        this.offsetX = 0;
        this.offsetY = 0;
    }
    
    /** Returns the widh of the box representing the agent
     * @return Widht in pixels of the agent
     */
    public int getWidth() {
        // calculates the width in pixels of the box from the width of the label
        if (this.manager.fontMetrics != null) {
            return(this.manager.fontMetrics.stringWidth(this.label) + 4 + 1);
        }
        return(5);
    }
    
    /** Returns the height of the box representing the agent
     * @return Height in pixels of the agent
     */
    public int getHeight() {
        // calculates the width in pixels of the box from the height of the label
        if (this.manager.fontMetrics != null) {
            return(this.manager.fontMetrics.getHeight() + 4 + 1);
        }
        return(5);
    }
    
    /** Paints the Agent on the graphics given in parameter
     * @param the graphics to draw the agent on
     */
    void paint(java.awt.Graphics gr) {
        int w = this.getWidth();
        int h = this.getHeight();
        // if the agent is linked with the one under the mouse, it will be drawn white
        if (this.highlight && this.manager.highlightRelated) {
            gr.setColor(Color.white);
        } else {
            gr.setColor(this.color);
        }
        // box and edge
        gr.fillRect(this.getLeft(), this.getTop(), w, h);
        gr.setColor(this.BOX_EDGE_COLOR);
        gr.drawRect(this.getLeft(), this.getTop(), w, h);
        // if the agent is well placed it the text is white
        if (this.wellPlaced && this.manager.highlightWellPlaced)
        {
            gr.setColor(Color.white);
        }
        // label
        gr.drawString(this.label, this.getLeft() + 2, this.getTop() + 2 + gr.getFontMetrics().getAscent());
    }
    
}
