/*
 * Link.java
 *
 * Created on April 24, 2002, 2:52 PM
 */

package epe.observationTools.graphLayoutTool.links;

import java.awt.Color;

import epe.observationTools.graphLayoutTool.agents.Agent;

/**
 * Internal representation of a Link
 * @author  Bastien Hofmann
 */
public class Link {
    
    /* Agent creating the link */
    public final Agent source;

    /* Agent being contacted by the link */
    public final Agent target;
    
    /** Creates a new instance of Link */
    public Link(Agent src, Agent tg) {
        this.source = src;
        this.target = tg;
    }
    
    /** Draws a line with a gradient between two colors
     * @param Graphics to draw the line on
     * @param x coordinate of the first point
     * @param y coordinate of the first point
     * @param x coordinate of the second point
     * @param y coordinate of the second point
     * @param color of the start of the gradient
     * @param color of the end of the gradient
     */
    void drawGradientLine(java.awt.Graphics gr, int x1, int y1, int x2, int y2, Color c1, Color c2) {
        int x = x1, y = y1, D = 0, HX = x2 - x1, HY = y2 - y1,
            c, M, xInc = 1, yInc = 1;
        
        gr.setColor(c1);
        float rBase = c1.getRed(),
            gBase = c1.getGreen(),
            bBase = c1.getBlue(),
            rStep, gStep, bStep;
        
        if (HX < 0){xInc = -1; HX = -HX;}
        if (HY < 0){yInc = -1; HY = -HY;}
        if (HY <= HX) {
            // step for the gradient
            rStep = (c2.getRed() - rBase) / HX;
            gStep = (c2.getGreen() - gBase) / HX;
            bStep = (c2.getBlue() - bBase) / HX;
            // line draw
            c = 2 * HX; M = 2 * HY;
            for (;;) {
                gr.drawLine(x, y, x, y);
                if (x == x2) break;
                x += xInc;
                D += M;
                if (D > HX) {
                    y += yInc; D -= c;
                }
                // changing color
                rBase += rStep;
                gBase += gStep;
                bBase += bStep;
                gr.setColor(new Color((int)rBase, (int)gBase, (int)bBase));
            }
        }
        else
        {
            // step for the gradient
            rStep = (c2.getRed() - rBase) / HY;
            gStep = (c2.getGreen() - gBase) / HY;
            bStep = (c2.getBlue() - bBase) / HY;
            // line draw
            c = 2 * HY; M = 2 * HX;
            for (;;) {
                gr.drawLine(x, y, x, y);
                if (y == y2) break;
                y += yInc;
                D += M;
                if (D > HY) {
                    x += xInc; D -= c;
                }
                // changing color
                rBase += rStep;
                gBase += gStep;
                bBase += bStep;
                gr.setColor(new Color((int)rBase, (int)gBase, (int)bBase));
            }
        }
    }

    /* Offset of the dot to animate the line */
    private int dotOffset = 0;
    
    /* size of the dot in pixels */
    private static final int DOT_SIZE = 3;
    
    /** Draws a dotted line with two colors
     * @param Graphics to draw the line on
     * @param x coordinate of the first point
     * @param y coordinate of the first point
     * @param x coordinate of the second point
     * @param y coordinate of the second point
     * @param color of the start of the gradient
     * @param color of the end of the gradient
     */
    void drawDotLine(java.awt.Graphics gr, int x1, int y1, int x2, int y2, Color c1, Color c2, boolean animate) {
        int x = x1, y = y1, D = 0, HX = x2 - x1, HY = y2 - y1,
            c, M, xInc = 1, yInc = 1, dotCounter = this.dotOffset;
        
        if (HX < 0){xInc = -1; HX = -HX;}
        if (HY < 0){yInc = -1; HY = -HY;}
        if (HY <= HX) {
            c = 2 * HX; M = 2 * HY;
            for (;;) {
                // color choice
                if (dotCounter >= Link.DOT_SIZE) {
                    gr.setColor(c2);
                } else {
                    gr.setColor(c1);
                }
                dotCounter = (dotCounter + 1) % (Link.DOT_SIZE * 2);
                gr.drawLine(x, y, x, y);
                if (x == x2) break;
                x += xInc;
                D += M;
                if (D > HX) {
                    y += yInc; D -= c;
                }
            }
        }
        else
        {
            c = 2 * HY; M = 2 * HX;
            for (;;) {
                // color choice
                if (dotCounter >= Link.DOT_SIZE) {
                    gr.setColor(c2);
                } else {
                    gr.setColor(c1);
                }
                dotCounter = (dotCounter + 1) % (Link.DOT_SIZE * 2);
                gr.drawLine(x, y, x, y);
                if (y == y2) break;
                y += yInc;
                D += M;
                if (D > HY) {
                    x += xInc; D -= c;
                }
            }
        }
        if (animate) {
            this.dotOffset = (this.dotOffset - 1) % (Link.DOT_SIZE * 2);
        }
    }
    
    /** Paints the link on the image given in parameter
     * @param the image to draw the link on
     */
    void paint(java.awt.Graphics gr, boolean animate) {
        //Graphics gr = img.getGraphics();
        // calculates the boundaries of the two agents having this link
        int l1 = this.source.getLeft();
        int t1 = this.source.getTop();
        int w1 = this.source.getWidth();
        int h1 = this.source.getHeight();
        int l2 = this.target.getLeft();
        int t2 = this.target.getTop();
        //int w2 = this.target.getWidth();
        int h2 = this.target.getHeight();
        // calculates the origin and the aim of the link
        int x1 = l1 + w1;
        int y1 = t1 + h1/2;
        int x2 = l2;
        int y2 = t2 + h2/2;
        /*if (Agent.DRAW_IO) {
            x1 += Agent.OUTPUT_WIDTH;
            x2 -= Agent.INPUT_SIZE;
        }*/
        // and draws it with a dotted line
        this.drawDotLine(gr, x1, y1, x2, y2, this.source.color, Color.black, animate);
    }
    
}
