/*
 * LinkManager.java
 *
 * Created on April 24, 2002, 3:41 PM
 */

package epe.observationTools.graphLayoutTool.links;

import java.util.Observer;

import epe.observationTools.graphLayoutTool.agents.Agent;
import epe.observationTools.graphLayoutTool.agents.AgentManager;
import epe.observationTools.graphLayoutTool.agents.AgentVector;

/**
 *
 * @author  Bastien Hofmann
 */
public class LinkManager implements Observer {
    
    /** Links managed by this LinkManager */
    public final LinkVector links;
    
    /** AgentManager having the agents concerned by these links */
    private final AgentManager agentManager;
    
    /** Creates a new instance of LinkManager */
    public LinkManager(AgentManager am) {
        this.agentManager = am;
        this.links = new LinkVector();
    }
    
    /** Adds a new link to the LinkManager with the specified Agents
     * @param source of the link
     * @param target of the link
     * @return true if the link doesn't already exist and could be created
     */
    public boolean add(Agent source, Agent target) {
        // if the link doesn't exists it is created
        if (this.links.indexOf(source, target)==-1) {
            return(this.links.add(new Link(source, target)));
        }
        // the link already exists
        return(false);
    }
    
    /** Adds a new link to the LinkManager with the specified labels
     * @param source of the link
     * @param target of the link
     * @return true if the link doesn't already exist and if the agents exist
     */
    public boolean add(java.lang.String source, java.lang.String target) {
        // looks for the agents in the agent manager
        int srcIndex = this.agentManager.agents.indexOf(source);
        int tgIndex = this.agentManager.agents.indexOf(target);
        if (srcIndex > -1 && tgIndex > -1) {
            // adds the link
            Agent src = this.agentManager.get(srcIndex);
            Agent tg = this.agentManager.get(tgIndex);
            return(this.add(src, tg));
        }
        // the specified agents have not been found
        return(false);
    }
    
    /** Removes the specified link from the LinkManager
     * @param source of the link to be removed
     * @param target of the link to be removed
     * @return true if the link could be removed
     */
    public boolean remove(Agent source, Agent target) {
        // looks for the link in the LinkManager
        int index = this.links.indexOf(source, target);
        // if it is found it is removed
        if (index>-1) {
            this.links.remove(index);
            return(true);
        }
        // the link has not been found
        return(false);
    }
    
    /** Removes the link specified by the labels given in argument from the LinkManager
     * @param source label
     * @param target label
     */
    public boolean remove(java.lang.String source, java.lang.String target) {
        // looks for the agents in the agent manager
        int srcIndex = this.agentManager.agents.indexOf(source);
        int tgIndex = this.agentManager.agents.indexOf(target);
        if (srcIndex>-1 && tgIndex>-1) {
            // removes the link
            Agent src = this.agentManager.get(srcIndex);
            Agent tg = this.agentManager.get(tgIndex);
            return(this.remove(src,tg));
        }
        // the agents have not been found
        return(false);
    }
    
    /** Removes all the links involving the agent given in parameter
     * @param Agent to remove the associated links
     */
    public void removeLinksFromAgent(Agent ag) {
        int i = 0;
        // looks for every link if it has the specified agent as target or source
        while (i < this.links.size()) {
            if (((Link)this.links.get(i)).source == ag || ((Link)this.links.get(i)).target == ag) {
                this.links.remove(i);
            } else {
                i++;
            }
        }
    }
    
    /** Returns an AgentVector containing the Agents being the sources
     * of links arriving to the agent specified
     * @param Agent to find the sources
     * @return AgentVector of the sources
     */
    public AgentVector getSources(Agent ag) {
        AgentVector av = new AgentVector();
        
        for (int i = 0; i < this.links.size(); i++) {
            if (((Link)this.links.get(i)).target == ag) {
                av.add(((Link)this.links.get(i)).source);
            }
        }
        return(av);
    }

    /** Looks if there's a link starting from the agent specified
     * @param The agent to be the source of a link
     * @return True if at least a link exists with this agent as source
     */
    public boolean isSource(Agent ag) {
        return(this.getFirstLinkAsSource(ag)>-1);
    }
    
    /** Returns the count of link for the specified agent as source
     * @param Agent to count the links
     * @return Link count
     */
    public int linksAsSource(Agent ag) {
        int c = 0;
        for (int i = 0; i < this.links.size(); i++) {
            if (((Link)(this.links.get(i))).source == ag) {
                c++;
            }
        }
        return(c);
    }
    
    /** Returns the index of the first link having the agent as source
     * @param Agent to look the link for
     * @return the index of the first link with agent as the source;
     * @return -1 otherwise
     */
    public int getFirstLinkAsSource(Agent ag) {
        // looks for every link the find the first one with ag as source
        for (int i = 0; i < this.links.size(); i++) {
            if (((Link)this.links.get(i)).source == ag) {
                return(i);
            }
        }
        return(-1);
    }
    
    /** Returns an AgentVector containing the Agents being the targets
     * of links initiated be the agent specified
     * @param Agent to find the targets
     * @return AgentVector of the targets
     */
    public AgentVector getTargets(Agent ag) {
        AgentVector av = new AgentVector();
        
        for (int i = 0; i < this.links.size(); i++) {
            if (((Link)this.links.get(i)).source == ag) {
                av.add(((Link)this.links.get(i)).target);
            }
        }
        return(av);
    }
    
    /** Looks if there's a link arriving to the specified agent
     * @param The agent supposed to be the target of a link
     * @return True if there's at least one link arriving to the agent
     */
    public boolean isTarget(Agent ag) {
        return(this.getFirstLinkAsTarget(ag)>-1);
    }
    
    /** Returns the index of the first link having the agent as target
     * @param Agent to look the link for
     * @return the index of the first link with agent as the target;
     * @return -1 otherwise
     */
    public int getFirstLinkAsTarget(Agent ag) {
        // looks for every link the find the first one with ag as target
        for (int i = 0; i < this.links.size(); i++) {
            if (((Link)this.links.get(i)).target == ag) {
                return(i);
            }
        }
        return(-1);
    }            
    
    /** Sets the agents being pointed be the mouse
     * @param Agent being pointed
     */
    public void setPointed(Agent ag) {
        if (AgentManager.pointed != ag) {
            AgentManager.pointed = ag;
            if (ag == null) {
                for (int i = 0; i < this.agentManager.agents.size(); i++) {
                    this.agentManager.get(i).highlight = false;
                }
            } else {
                for (int i = 0; i < this.agentManager.agents.size(); i++) {
                    Agent a = this.agentManager.get(i);
                    if ((this.links.indexOf(AgentManager.pointed, a) > -1) ||
                        (this.links.indexOf(a, AgentManager.pointed) > -1)) {
                            a.highlight = true;
                    } else {
                        a.highlight = false;
                    }
                }
            }
        }
    }
    
    /** Paints the links on the image given in parameter
     * @param image to draw the links on
     */
    public void paint(java.awt.Graphics gr, boolean animate) {
        // paints the links one by one
        for (int i = 0; i < this.links.size(); i++) {
            this.links.link(i).paint(gr, animate);
        }
    }
    
    public void update(java.util.Observable observable, Object obj) {
        this.removeLinksFromAgent((Agent)obj);
    }
    
}
